from abc import ABC, abstractmethod
from collections.abc import Iterable, Iterator
from dataclasses import dataclass

from astroid import Attribute, Call, Const, InferenceError, NodeNG  # type: ignore
from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.source_code.base import (
    Advice,
    Advisory,
    Deprecation,
    Fixer,
    CurrentSessionState,
    PythonLinter,
)
from databricks.labs.ucx.source_code.linters.python_infer import InferredValue
from databricks.labs.ucx.source_code.queries import FromTableSqlLinter
from databricks.labs.ucx.source_code.linters.python_ast import Tree, TreeHelper


@dataclass
class Matcher(ABC):
    method_name: str
    min_args: int
    max_args: int
    table_arg_index: int
    table_arg_name: str | None = None
    call_context: dict[str, set[str]] | None = None
    session_state: CurrentSessionState | None = None

    def matches(self, node: NodeNG):
        return (
            isinstance(node, Call)
            and self._get_table_arg(node) is not None
            and isinstance(node.func, Attribute)
            and Tree(node.func.expr).is_from_module("spark")
        )

    @abstractmethod
    def lint(
        self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state: CurrentSessionState, node: Call
    ) -> Iterator[Advice]:
        """raises Advices by linting the code"""

    @abstractmethod
    def apply(self, from_table: FromTableSqlLinter, index: MigrationIndex, node: Call) -> None:
        """applies recommendations"""

    def _get_table_arg(self, node: Call):
        node_argc = len(node.args)
        if self.min_args <= node_argc <= self.max_args and self.table_arg_index < node_argc:
            return node.args[self.table_arg_index]
        if not self.table_arg_name:
            return None
        if not node.keywords:
            return None
        for keyword in node.keywords:
            if keyword.arg == self.table_arg_name:
                return keyword.value
        return None

    def _check_call_context(self, node: Call) -> bool:
        assert isinstance(node.func, Attribute)  # Avoid linter warning
        func_name = node.func.attrname
        qualified_name = TreeHelper.get_full_function_name(node)

        # Check if the call_context is None as that means all calls are checked
        if self.call_context is None:
            return True

        # Get the qualified names from the call_context dictionary
        qualified_names = self.call_context.get(func_name)

        # Check if the qualified name is in the set of qualified names that are allowed
        return qualified_name in qualified_names if qualified_names else False


@dataclass
class QueryMatcher(Matcher):

    def lint(
        self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state: CurrentSessionState, node: Call
    ) -> Iterator[Advice]:
        table_arg = self._get_table_arg(node)
        if table_arg:
            try:
                for inferred in InferredValue.infer_from_node(table_arg, self.session_state):
                    yield from self._lint_table_arg(from_table, node, inferred)
            except InferenceError:
                yield Advisory.from_node(
                    code='cannot-autofix-table-reference',
                    message=f"Can't migrate table_name argument in '{node.as_string()}' because its value cannot be computed",
                    node=node,
                )

    @classmethod
    def _lint_table_arg(cls, from_table: FromTableSqlLinter, call_node: NodeNG, inferred: InferredValue):
        if inferred.is_inferred():
            for advice in from_table.lint(inferred.as_string()):
                yield advice.replace_from_node(call_node)
        else:
            yield Advisory.from_node(
                code='cannot-autofix-table-reference',
                message=f"Can't migrate table_name argument in '{call_node.as_string()}' because its value cannot be computed",
                node=call_node,
            )

    def apply(self, from_table: FromTableSqlLinter, index: MigrationIndex, node: Call) -> None:
        table_arg = self._get_table_arg(node)
        assert isinstance(table_arg, Const)
        new_query = from_table.apply(table_arg.value)
        table_arg.value = new_query


@dataclass
class TableNameMatcher(Matcher):

    def lint(
        self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state: CurrentSessionState, node: Call
    ) -> Iterator[Advice]:
        table_arg = self._get_table_arg(node)
        table_name = table_arg.as_string().strip("'").strip('"')
        for inferred in InferredValue.infer_from_node(table_arg, session_state):
            if not inferred.is_inferred():
                yield Advisory.from_node(
                    code='cannot-autofix-table-reference',
                    message=f"Can't migrate '{node.as_string()}' because its table name argument cannot be computed",
                    node=node,
                )
                continue
            dst = self._find_dest(index, inferred.as_string(), from_table.schema)
            if dst is None:
                continue
            yield Deprecation.from_node(
                code='table-migrated-to-uc',
                message=f"Table {table_name} is migrated to {dst.destination()} in Unity Catalog",
                # SQLGlot does not propagate tokens yet. See https://github.com/tobymao/sqlglot/issues/3159
                node=node,
            )

    def apply(self, from_table: FromTableSqlLinter, index: MigrationIndex, node: Call) -> None:
        table_arg = self._get_table_arg(node)
        assert isinstance(table_arg, Const)
        dst = self._find_dest(index, table_arg.value, from_table.schema)
        if dst is not None:
            table_arg.value = dst.destination()

    @staticmethod
    def _find_dest(index: MigrationIndex, value: str, schema: str):
        parts = value.split(".")
        # Ensure that unqualified table references use the current schema
        if len(parts) == 1:
            return index.get(schema, parts[0])
        return None if len(parts) != 2 else index.get(parts[0], parts[1])


@dataclass
class ReturnValueMatcher(Matcher):

    def matches(self, node: NodeNG):
        return (
            isinstance(node, Call) and isinstance(node.func, Attribute) and Tree(node.func.expr).is_from_module("spark")
        )

    def lint(
        self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state: CurrentSessionState, node: Call
    ) -> Iterator[Advice]:
        assert isinstance(node.func, Attribute)  # always true, avoids a pylint warning
        yield Advisory.from_node(
            code='changed-result-format-in-uc',
            message=f"Call to '{node.func.attrname}' will return a list of <catalog>.<database>.<table> instead of <database>.<table>.",
            node=node,
        )

    def apply(self, from_table: FromTableSqlLinter, index: MigrationIndex, node: Call) -> None:
        # No transformations to apply
        return


@dataclass
class DirectFilesystemAccessMatcher(Matcher):
    _DIRECT_FS_REFS = {
        "s3a://",
        "s3n://",
        "s3://",
        "wasb://",
        "wasbs://",
        "abfs://",
        "abfss://",
        "dbfs:/",
        "hdfs://",
        "file:/",
    }

    def matches(self, node: NodeNG):
        return (
            isinstance(node, Call)
            and self._get_table_arg(node) is not None
            and isinstance(node.func, Attribute)
            and (Tree(node.func.expr).is_from_module("spark") or Tree(node.func.expr).is_from_module("dbutils"))
        )

    def lint(
        self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state: CurrentSessionState, node: NodeNG
    ) -> Iterator[Advice]:
        table_arg = self._get_table_arg(node)
        if not isinstance(table_arg, Const):
            return
        if not table_arg.value:
            return
        if not isinstance(table_arg.value, str):
            return
        if any(table_arg.value.startswith(prefix) for prefix in self._DIRECT_FS_REFS):
            yield Deprecation.from_node(
                code='direct-filesystem-access',
                message=f"The use of direct filesystem references is deprecated: {table_arg.value}",
                node=node,
            )
            return
        if table_arg.value.startswith("/") and self._check_call_context(node):
            yield Deprecation.from_node(
                code='implicit-dbfs-usage',
                message=f"The use of default dbfs: references is deprecated: {table_arg.value}",
                node=node,
            )

    def apply(self, from_table: FromTableSqlLinter, index: MigrationIndex, node: Call) -> None:
        # No transformations to apply
        return


class SparkMatchers:

    def __init__(self):
        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.SparkSession.html
        spark_session_matchers = [QueryMatcher("sql", 1, 1000, 0, "sqlQuery"), TableNameMatcher("table", 1, 1, 0)]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Catalog.html
        spark_catalog_matchers = [
            TableNameMatcher("cacheTable", 1, 2, 0, "tableName"),
            TableNameMatcher("createTable", 1, 1000, 0, "tableName"),
            TableNameMatcher("createExternalTable", 1, 1000, 0, "tableName"),
            TableNameMatcher("getTable", 1, 1, 0),
            TableNameMatcher("isCached", 1, 1, 0),
            TableNameMatcher("listColumns", 1, 2, 0, "tableName"),
            TableNameMatcher("tableExists", 1, 2, 0, "tableName"),
            TableNameMatcher("recoverPartitions", 1, 1, 0),
            TableNameMatcher("refreshTable", 1, 1, 0),
            TableNameMatcher("uncacheTable", 1, 1, 0),
            ReturnValueMatcher("listTables", 0, 2, 0),
        ]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.html
        spark_dataframe_matchers = [
            TableNameMatcher("writeTo", 1, 1, 0),
        ]

        # nothing to migrate in Column, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Column.html
        # nothing to migrate in Observation, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Observation.html
        # nothing to migrate in Row, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Row.html
        # nothing to migrate in GroupedData, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.GroupedData.html
        # nothing to migrate in PandasCogroupedOps, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.PandasCogroupedOps.html
        # nothing to migrate in DataFrameNaFunctions, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameNaFunctions.html
        # nothing to migrate in DataFrameStatFunctions, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameStatFunctions.html
        # nothing to migrate in Window, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.Window.html

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameReader.html
        spark_dataframereader_matchers = [
            TableNameMatcher("table", 1, 1, 0),  # TODO good example of collision, see spark_session_calls
        ]

        # see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameWriter.html
        spark_dataframewriter_matchers = [
            TableNameMatcher("insertInto", 1, 2, 0, "tableName"),
            # TODO jdbc: could the url be a databricks url, raise warning ?
            TableNameMatcher("saveAsTable", 1, 4, 0, "name"),
        ]

        # nothing to migrate in DataFrameWriterV2, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrameWriterV2.html
        # nothing to migrate in UDFRegistration, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UDFRegistration.html

        direct_fs_access_matchers = [
            DirectFilesystemAccessMatcher("ls", 1, 1, 0, call_context={"ls": {"dbutils.fs.ls"}}),
            DirectFilesystemAccessMatcher("cp", 1, 2, 0, call_context={"cp": {"dbutils.fs.cp"}}),
            DirectFilesystemAccessMatcher("rm", 1, 1, 0, call_context={"rm": {"dbutils.fs.rm"}}),
            DirectFilesystemAccessMatcher("head", 1, 1, 0, call_context={"head": {"dbutils.fs.head"}}),
            DirectFilesystemAccessMatcher("put", 1, 2, 0, call_context={"put": {"dbutils.fs.put"}}),
            DirectFilesystemAccessMatcher("mkdirs", 1, 1, 0, call_context={"mkdirs": {"dbutils.fs.mkdirs"}}),
            DirectFilesystemAccessMatcher("mv", 1, 2, 0, call_context={"mv": {"dbutils.fs.mv"}}),
            DirectFilesystemAccessMatcher("text", 1, 3, 0),
            DirectFilesystemAccessMatcher("csv", 1, 1000, 0),
            DirectFilesystemAccessMatcher("json", 1, 1000, 0),
            DirectFilesystemAccessMatcher("orc", 1, 1000, 0),
            DirectFilesystemAccessMatcher("parquet", 1, 1000, 0),
            DirectFilesystemAccessMatcher("save", 0, 1000, 0, "path"),
            DirectFilesystemAccessMatcher("load", 0, 1000, 0, "path"),
            DirectFilesystemAccessMatcher("option", 1, 1000, 1),  # Only .option("path", "xxx://bucket/path") will hit
            DirectFilesystemAccessMatcher("addFile", 1, 3, 0),
            DirectFilesystemAccessMatcher("binaryFiles", 1, 2, 0),
            DirectFilesystemAccessMatcher("binaryRecords", 1, 2, 0),
            DirectFilesystemAccessMatcher("dump_profiles", 1, 1, 0),
            DirectFilesystemAccessMatcher("hadoopFile", 1, 8, 0),
            DirectFilesystemAccessMatcher("newAPIHadoopFile", 1, 8, 0),
            DirectFilesystemAccessMatcher("pickleFile", 1, 3, 0),
            DirectFilesystemAccessMatcher("saveAsHadoopFile", 1, 8, 0),
            DirectFilesystemAccessMatcher("saveAsNewAPIHadoopFile", 1, 7, 0),
            DirectFilesystemAccessMatcher("saveAsPickleFile", 1, 2, 0),
            DirectFilesystemAccessMatcher("saveAsSequenceFile", 1, 2, 0),
            DirectFilesystemAccessMatcher("saveAsTextFile", 1, 2, 0),
            DirectFilesystemAccessMatcher("load_from_path", 1, 1, 0),
        ]

        # nothing to migrate in UserDefinedFunction, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UserDefinedFunction.html
        # nothing to migrate in UserDefinedTableFunction, see https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.UserDefinedTableFunction.html
        self._matchers = {}
        for matcher in (
            spark_session_matchers
            + spark_catalog_matchers
            + spark_dataframe_matchers
            + spark_dataframereader_matchers
            + spark_dataframewriter_matchers
            + direct_fs_access_matchers
        ):
            self._matchers[matcher.method_name] = matcher

    @property
    def matchers(self):
        return self._matchers


class SparkSqlPyLinter(PythonLinter, Fixer):

    _spark_matchers = SparkMatchers()

    def __init__(self, from_table: FromTableSqlLinter, index: MigrationIndex, session_state):
        self._from_table = from_table
        self._index = index
        self._session_state = session_state

    def name(self) -> str:
        # this is the same fixer, just in a different language context
        return self._from_table.name()

    def lint_tree(self, tree: Tree) -> Iterable[Advice]:
        for node in tree.walk():
            matcher = self._find_matcher(node)
            if matcher is None:
                continue
            assert isinstance(node, Call)
            yield from matcher.lint(self._from_table, self._index, self._session_state, node)

    def apply(self, code: str) -> str:
        tree = Tree.parse(code)
        # we won't be doing it like this in production, but for the sake of the example
        for node in tree.walk():
            matcher = self._find_matcher(node)
            if matcher is None:
                continue
            assert isinstance(node, Call)
            matcher.apply(self._from_table, self._index, node)
        return tree.node.as_string()

    def _find_matcher(self, node: NodeNG):
        if not isinstance(node, Call):
            return None
        if not isinstance(node.func, Attribute):
            return None
        matcher = self._spark_matchers.matchers.get(node.func.attrname, None)
        if matcher is None:
            return None
        return matcher if matcher.matches(node) else None



import logging
from collections.abc import Iterable
from dataclasses import dataclass, replace
from functools import partial

from databricks.labs.blueprint.parallel import Threads
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk.errors import Unknown, NotFound

from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier

logger = logging.getLogger(__name__)


@dataclass
class Udf:  # pylint: disable=too-many-instance-attributes
    catalog: str
    database: str
    name: str
    func_type: str
    func_input: str
    func_returns: str
    deterministic: bool
    data_access: str
    body: str
    comment: str
    success: int = 1
    failures: str = ""

    @property
    def key(self) -> str:
        return f"{self.catalog}.{self.database}.{self.name}".lower()


class UdfsCrawler(CrawlerBase):
    def __init__(self, backend: SqlBackend, schema: str, include_databases: list[str] | None = None):
        """
        Initializes a UdfsCrawler instance.

        Args:
            backend (SqlBackend): The SQL Execution Backend abstraction (either REST API or Spark)
            schema: The schema name for the inventory persistence.
        """
        super().__init__(backend, "hive_metastore", schema, "udfs", Udf)
        self._include_database = include_databases

    def _all_databases(self) -> list[str]:
        if not self._include_database:
            return [row[0] for row in self._fetch("SHOW DATABASES")]
        return self._include_database

    def snapshot(self) -> list[Udf]:
        """
        Takes a snapshot of tables in the specified catalog and database.

        Returns:
            list[Udf]: A list of Udf objects representing the snapshot of tables.
        """
        return self._snapshot(self._try_load, self._crawl)

    def _try_load(self) -> Iterable[Udf]:
        """Tries to load udf information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Udf(*row)

    def _crawl(self) -> Iterable[Udf]:
        """Crawls and lists udfs within the specified catalog and database."""
        tasks = []
        catalog = "hive_metastore"
        # need to set the current catalog otherwise "SHOW USER FUNCTIONS FROM" is raising error:
        # "target schema <database> is not in the current catalog"
        self._exec(f"USE CATALOG {escape_sql_identifier(catalog)};")
        for database in self._all_databases():
            for task in self._collect_tasks(catalog, database):
                tasks.append(task)
        udfs, errors = Threads.gather(f"listing udfs in {catalog}", tasks)
        if len(errors) > 0:
            logger.error(f"Detected {len(errors)} while scanning udfs in {catalog}")
        return self._assess_udfs(udfs)

    def _collect_tasks(self, catalog, database) -> Iterable[partial[Udf | None]]:
        """
        Hive metastore supports 2 type of UDFs, SQL UDFs & Hive UDFs, both are created using CREATE FUNCTION
        - https://docs.databricks.com/en/udf/index.html
        - https://docs.databricks.com/en/sql/language-manual/sql-ref-functions-udf-hive.html
        We do not have to worry about Python or Scala UDF/UDAFs, as they cannot be permanent registered in HMS. Those
        UDFs when defined are only valid within the Spark session.
        Unity Catalog supports SQL UDFs, Python UDFs, which means SQL UDFs can be migrated as is. Hive UDFs cannot be
        migrated
        """
        try:
            logger.debug(f"[{catalog}.{database}] listing udfs")
            for (udf,) in self._fetch(
                f"SHOW USER FUNCTIONS FROM {escape_sql_identifier(catalog)}.{escape_sql_identifier(database)};"
            ):
                if not udf.startswith(f"{catalog}.{database}"):
                    continue
                udf_name = udf[udf.rfind(".") + 1 :]  # remove catalog and database info from the name
                yield partial(self._describe, catalog, database, udf_name)
        except NotFound:
            # This make the integration test more robust as many test schemas are being created and deleted quickly.
            logger.warning(f"Schema {catalog}.{database} no longer existed")
        except Unknown as err:
            logger.error(f"Problem with {database}: {err}")

    def _describe(self, catalog: str, database: str, udf: str) -> Udf | None:
        """Fetches metadata like udf type, input, returns, data access and body
        if specified for a specific udf within the given catalog and database.
        The output is different between SQL UDFs and Hive UDFs. Hive UDFs only returns the class & usage.
        """
        full_name = f"{catalog}.{database}.{udf}"
        logger.debug(f"[{full_name}] fetching udf metadata")
        describe = {}
        current_key = ""
        try:
            for row in self._fetch(f"DESCRIBE FUNCTION EXTENDED {escape_sql_identifier(full_name)}"):
                key_value = row["function_desc"]
                if ":" in key_value:
                    current_key, value = key_value.split(":", 1)
                    describe[current_key] = value.strip()
                elif current_key != "":  # append multiline returns, e.g. Input
                    describe[current_key] += f"\n{key_value.strip()}"
            return Udf(
                catalog=catalog.lower(),
                database=database.lower(),
                name=udf.lower(),
                func_type=describe.get("Type", describe.get("Class", "UNKNOWN")),
                func_input=describe.get("Input", "UNKNOWN"),
                func_returns=describe.get("Returns", "UNKNOWN"),
                deterministic=bool(describe.get("Deterministic", False)),
                data_access=describe.get("Data Access", "UNKNOWN"),
                comment=describe.get("Comment", "UNKNOWN"),
                body=describe.get("Body", "UNKNOWN"),
            )
        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error(f"Couldn't fetch information for udf {full_name} : {e}")
            return None

    @staticmethod
    def _assess_udfs(udfs: Iterable[Udf]) -> Iterable[Udf]:
        for udf in udfs:
            if udf.func_type != "SCALAR":
                yield replace(udf, success=0, failures="Only SCALAR functions are supported")
            else:
                yield replace(udf, success=1)



import logging
from collections.abc import Collection
from dataclasses import dataclass
from functools import cached_property

import sqlglot
from sqlglot import ParseError, expressions

from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex, TableView
from databricks.labs.ucx.hive_metastore.mapping import TableToMigrate
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.queries import FromTableSqlLinter

logger = logging.getLogger(__name__)


@dataclass
class ViewToMigrate(TableToMigrate):
    def __post_init__(self):
        if self.src.view_text is None:
            raise RuntimeError("Should never get there! A view must have 'view_text'!")

    @cached_property
    def dependencies(self) -> list[TableView]:
        return list(self._view_dependencies())

    def _view_dependencies(self):
        try:
            statements = sqlglot.parse(self.src.view_text, read='databricks')
        except ParseError as e:
            raise ValueError(f"Could not analyze view SQL: {self.src.view_text}") from e
        if len(statements) != 1 or statements[0] is None:
            raise ValueError(f"Could not analyze view SQL: {self.src.view_text}")
        statement = statements[0]
        aliases = self._read_aliases(statement)
        for old_table in statement.find_all(expressions.Table):
            if old_table.name in aliases:
                continue
            if old_table.catalog and old_table.catalog != 'hive_metastore':
                continue
            src_db = old_table.db if old_table.db else self.src.database
            if not src_db:
                logger.error(f"Could not determine schema for table {old_table.name}")
                continue
            yield TableView("hive_metastore", src_db, old_table.name)

    def _read_aliases(self, statement: expressions.Expression):
        aliases = set()
        for with_clause in statement.find_all(expressions.With):
            for expression in with_clause.expressions:
                if isinstance(expression, expressions.CTE):
                    aliases.add(expression.alias_or_name)
        return aliases

    def sql_migrate_view(self, index: MigrationIndex) -> str:
        from_table = FromTableSqlLinter(index, CurrentSessionState(self.src.database))
        assert self.src.view_text is not None, 'Expected a view text'
        migrated_select = from_table.apply(self.src.view_text)
        statements = sqlglot.parse(migrated_select, read='databricks')
        assert len(statements) == 1, 'Expected a single statement'
        create = statements[0]
        assert isinstance(create, expressions.Create), 'Expected a CREATE statement'
        # safely replace current table name with the updated catalog
        for table_name in create.find_all(expressions.Table):
            if table_name.db == self.src.database and table_name.name == self.src.name:
                # See https://github.com/tobymao/sqlglot/issues/3311
                new_view_name = expressions.Table(
                    catalog=expressions.to_identifier(self.rule.catalog_name),
                    db=expressions.to_identifier(self.rule.dst_schema),
                    this=expressions.to_identifier(self.rule.dst_table),
                )
                table_name.replace(new_view_name)
        # safely replace CREATE with CREATE IF NOT EXISTS
        create.args['exists'] = True
        return create.sql('databricks', identify=True)

    def __hash__(self):
        return hash(self.src)

    def __eq__(self, other):
        return isinstance(other, TableToMigrate) and self.src == other.src


class ViewsMigrationSequencer:

    def __init__(self, tables_to_migrate: Collection[TableToMigrate], *, migration_index: MigrationIndex | None = None):
        self._tables = tables_to_migrate  # Also contains views to migrate
        self._index = migration_index or MigrationIndex([])

    @cached_property
    def _views(self) -> dict[ViewToMigrate, TableView]:
        # Views is a mapping as the TableView is required when resolving dependencies
        views = {}
        for table_or_view in self._tables:
            if table_or_view.src.view_text is None:
                continue
            view_to_migrate = ViewToMigrate(table_or_view.src, table_or_view.rule)
            # All views to migrate are stored in the hive_metastore
            views[view_to_migrate] = TableView("hive_metastore", view_to_migrate.src.database, view_to_migrate.src.name)
        return views

    def _get_view_to_migrate(self, key: str) -> ViewToMigrate | None:
        """Get a view to migrate by key"""
        for view in self._views:
            if view.src.key == key:
                return view
        return None

    def sequence_batches(self) -> list[list[ViewToMigrate]]:
        """Sequence the views in batches to migrate them in the right order.

        Batch sequencing uses the following algorithm:
        0. For each view, we register dependencies (extracted from view_text),
        1. Then to create a new batch of views,
           We require the dependencies that are covered already:
             1. The migrated tables
             2. The (growing) set of views from already sequenced previous batches
           For each remaining view, we check if all its dependencies are covered for. If that is the case, then we
           add that view to the new batch of views.
        2. We repeat point from point 1. until all views are sequenced.

        The complexity for a given set of views v and a dependency depth d looks like Ov^d, this seems enormous but in
        practice d remains small and v decreases rapidly
        """
        batches: list[list[ViewToMigrate]] = []
        views_to_migrate = set(self._views.keys())
        views_sequenced: set[TableView] = set()
        while len(views_to_migrate) > 0:
            try:
                next_batch = self._next_batch(views_to_migrate, views_from_previous_batches=views_sequenced)
            except RecursionError as e:
                logger.error(
                    f"Cannot sequence views {views_to_migrate} given migration index {self._index}", exc_info=e
                )
                # By returning the current batches, we can migrate the views that are not causing the recursion
                return batches
            for view in next_batch:
                views_sequenced.add(self._views[view])
            batches.append(next_batch)
            views_to_migrate.difference_update(next_batch)
        return batches

    def _next_batch(
        self, views: set[ViewToMigrate], *, views_from_previous_batches: set[TableView] | None
    ) -> list[ViewToMigrate]:
        """For sequencing algorithm see docstring of :meth:sequence_batches.

        Raises:
            RecursionError :
                If an infinite loop is detected.
        """
        views_from_previous_batches = views_from_previous_batches or set()
        # we can't (slightly) optimize by checking len(views) == 0 or 1,
        # because we'd lose the opportunity to check the SQL
        result = []
        for view in views:
            self._check_circular_dependency(view)
            if len(view.dependencies) == 0:
                result.append(view)
                continue
            # If all dependencies are already processed, we can add the view to the next batch
            not_processed_yet = set(view.dependencies) - views_from_previous_batches
            if len(not_processed_yet) == 0:
                result.append(view)
                continue
            if all(self._index.is_migrated(table_view.schema, table_view.name) for table_view in not_processed_yet):
                result.append(view)
        if len(result) == 0 and len(views) > 0:  # prevent infinite loop
            raise RecursionError(f"Unresolved dependencies prevent batch sequencing: {views}")
        return result

    def _check_circular_dependency(self, view: ViewToMigrate) -> None:
        """Check for circular dependencies in the views to migrate.

        Raises:
            RecursionError :
                If a circular dependency is detected between views.
        """
        dependency_keys = [dep.key for dep in view.dependencies]
        while len(dependency_keys) > 0:
            dependency = self._get_view_to_migrate(dependency_keys.pop(0))
            if not dependency:  # Only views (to migrate) can cause a circular dependency, tables can be ignored
                continue
            if dependency == view:
                raise RecursionError(f"Circular dependency detected starting from: {view.src.full_name}")
            dependency_keys.extend(dep.key for dep in dependency.dependencies)



import collections
import json
import logging
from collections.abc import Iterable, Callable
from dataclasses import dataclass
from functools import cached_property

from databricks.labs.lsql import Row
from databricks.sdk import WorkspaceClient

from databricks.labs.blueprint.installation import NotInstalled

from databricks.labs.ucx.account.workspaces import AccountWorkspaces
from databricks.labs.ucx.contexts.workspace_cli import WorkspaceContext
from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex, MigrationStatus
from databricks.labs.ucx.hive_metastore.locations import LocationTrie
from databricks.labs.ucx.hive_metastore.tables import Table
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.queries import FromTableSqlLinter

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@dataclass
class AssessmentObject:
    workspace_id: int
    object_type: str
    object_id: str
    failures: list[str]


@dataclass
class TableFromWorkspace(Table):
    workspace_id: int = 0

    @property
    def key(self) -> str:
        return f"{self.workspace_id}:{super().key}"


class AccountAggregate:
    def __init__(
        self,
        account_workspaces: AccountWorkspaces,
        workspace_context_factory: Callable[[WorkspaceClient], WorkspaceContext] = WorkspaceContext,
    ):
        self._account_workspaces = account_workspaces
        self._workspace_context_factory = workspace_context_factory

    @cached_property
    def _workspace_contexts(self):
        contexts = []
        for workspace_client in self._account_workspaces.workspace_clients():
            ctx = self._workspace_context_factory(workspace_client)
            contexts.append(ctx)
        return contexts

    def _federated_ucx_query(self, query: str, table_name='objects') -> Iterable[tuple[int, Row]]:
        """Modifies a query with a workspace-specific UCX schema and executes it on each workspace,
        yielding a tuple of workspace_id and Row. This means that you don't have to specify a database in the query,
        as it will be replaced with the UCX schema for each workspace. Use this method to aggregate results across
        all workspaces, where UCX is installed.

        At the moment, it's done sequentially, which is theoretically inefficient, but the number of workspaces is
        expected to be small. If this becomes a performance bottleneck, we can optimize it later via Threads.strict()
        """
        for ctx in self._workspace_contexts:
            workspace_id = ctx.workspace_client.get_workspace_id()
            try:
                # use already existing code to replace tables in the query, assuming that UCX database is in HMS
                inventory_database = ctx.config.inventory_database
                stub_index = MigrationIndex(
                    [
                        MigrationStatus(
                            src_schema=inventory_database,
                            src_table=table_name,
                            dst_catalog='hive_metastore',
                            dst_schema=inventory_database,
                            dst_table=table_name,
                        )
                    ]
                )
                from_table = FromTableSqlLinter(stub_index, CurrentSessionState(schema=inventory_database))
                logger.info(f"Querying Schema {inventory_database}")

                workspace_specific_query = from_table.apply(query)
                for row in ctx.sql_backend.fetch(workspace_specific_query):
                    yield workspace_id, row
            except NotInstalled:
                logger.warning(f"Workspace {workspace_id} does not have UCX installed")

    @cached_property
    def _aggregate_objects(self) -> list[AssessmentObject]:
        objects = []
        # view is defined in src/databricks/labs/ucx/queries/views/objects.sql
        for workspace_id, row in self._federated_ucx_query('SELECT * FROM objects'):
            objects.append(AssessmentObject(workspace_id, row.object_type, row.object_id, json.loads(row.failures)))
        return objects

    def _fetch_tables(self) -> Iterable[TableFromWorkspace]:
        for workspace_id, row in self._federated_ucx_query("SELECT * FROM tables", table_name="tables"):
            # Mypy complains about multiple values for `workspace_id`
            yield TableFromWorkspace(*row, workspace_id=workspace_id)  # type: ignore

    def readiness_report(self):
        logger.info("Generating readiness report")
        all_objects = 0
        incompatible_objects = 0
        failures = collections.defaultdict(list)

        for obj in self._aggregate_objects:
            all_objects += 1
            has_failures = False
            for failure in obj.failures:
                failures[failure].append(obj)
                has_failures = True
            if has_failures:
                incompatible_objects += 1
        compatibility = (1 - incompatible_objects / all_objects if all_objects > 0 else 0) * 100
        logger.info(f"UC compatibility: {compatibility}% ({incompatible_objects}/{all_objects})")

        for failure, objects in failures.items():
            logger.info(f"{failure}: {len(objects)} objects")

    def validate_table_locations(self) -> list[list[Table]]:
        """The table locations should not be overlapping."""
        logger.info("Validating migration readiness")
        tables = list(self._fetch_tables())
        trie = LocationTrie()
        for table in tables:
            if table.location is not None:
                trie.insert(table)
        seen_tables, all_conflicts = set(), []
        for table in tables:
            if table.key in seen_tables:
                continue
            if table.location is None:
                continue
            seen_tables.add(table.key)
            node = trie.find(table)
            if node is None:
                continue
            if not node.has_children() and len(node.tables) == 1:
                continue
            conflicts = []
            for sub_node in node:
                for conflicting_table in sub_node.tables:
                    conflicts.append(conflicting_table)
                    seen_tables.add(conflicting_table.key)
            conflict_message = " and ".join(conflict.key for conflict in conflicts)
            logger.warning(f"Overlapping table locations: {conflict_message}")
            all_conflicts.append(conflicts)
        return all_conflicts



from __future__ import annotations

import abc
import logging
from dataclasses import dataclass
from pathlib import Path
from collections.abc import Callable, Iterable, Iterator
from typing import TypeVar, Generic

from astroid import (  # type: ignore
    NodeNG,
)
from databricks.labs.ucx.source_code.base import Advisory, CurrentSessionState, is_a_notebook
from databricks.labs.ucx.source_code.linters.python_ast import Tree
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.Logger(__name__)


class DependencyGraph:

    def __init__(
        self,
        dependency: Dependency,
        parent: DependencyGraph | None,
        resolver: DependencyResolver,
        path_lookup: PathLookup,
        session_state: CurrentSessionState,
    ):
        self._dependency = dependency
        self._parent = parent
        self._resolver = resolver
        self._path_lookup = path_lookup.change_directory(dependency.path.parent)
        self._session_state = session_state
        self._dependencies: dict[Dependency, DependencyGraph] = {}

    @property
    def path_lookup(self):
        return self._path_lookup

    @property
    def dependency(self):
        return self._dependency

    def register_library(self, *libraries: str) -> list[DependencyProblem]:
        return self._resolver.register_library(self.path_lookup, *libraries)

    def register_notebook(self, path: Path, inherit_context: bool) -> list[DependencyProblem]:
        maybe = self._resolver.resolve_notebook(self.path_lookup, path, inherit_context)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_import(self, name: str) -> list[DependencyProblem]:
        if not name:
            return [DependencyProblem('import-empty', 'Empty import name')]
        maybe = self._resolver.resolve_import(self.path_lookup, name)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_file(self, path: Path) -> list[DependencyProblem]:
        maybe = self._resolver.resolve_file(self.path_lookup, path)
        if not maybe.dependency:
            return maybe.problems
        maybe_graph = self.register_dependency(maybe.dependency)
        return maybe_graph.problems

    def register_dependency(self, dependency: Dependency) -> MaybeGraph:
        # TODO: this has to be a private method, because we don't want to allow free-form dependencies.
        # the only case we have for this method to be used outside of this class is for DistInfoPackage
        # See databricks.labs.ucx.source_code.python_libraries.DistInfoContainer.build_dependency_graph for reference
        logger.debug(f"Registering dependency {dependency}")
        maybe = self.locate_dependency(dependency.path)
        if maybe.graph is not None:
            self._dependencies[dependency] = maybe.graph
            return maybe
        # nay, create the child graph and populate it
        child_graph = DependencyGraph(dependency, self, self._resolver, self._path_lookup, self._session_state)
        self._dependencies[dependency] = child_graph
        container = dependency.load(self.path_lookup)
        # TODO: Return either (child) graph OR problems
        if not container:
            problem = DependencyProblem('dependency-register-failed', 'Failed to register dependency', dependency.path)
            return MaybeGraph(child_graph, [problem])
        problems = container.build_dependency_graph(child_graph)
        return MaybeGraph(
            child_graph,
            [
                problem.replace(
                    source_path=dependency.path if problem.is_path_missing() else problem.source_path,
                )
                for problem in problems
            ],
        )

    def locate_dependency(self, path: Path) -> MaybeGraph:
        # need a list since unlike JS, Python won't let you assign closure variables
        found: list[DependencyGraph] = []

        def check_registered_dependency(graph):
            if graph.dependency.path == path:
                found.append(graph)
                return True
            return False

        self.root.visit(check_registered_dependency, set())
        if not found:
            return MaybeGraph(None, [DependencyProblem('dependency-not-found', 'Dependency not found')])
        return MaybeGraph(found[0], [])

    @property
    def root(self):
        return self if self._parent is None else self._parent.root

    @property
    def parent(self):
        return self._parent

    @property
    def all_dependencies(self) -> set[Dependency]:
        dependencies: set[Dependency] = set()

        def add_to_dependencies(graph: DependencyGraph) -> bool:
            if graph.dependency in dependencies:
                return True
            dependencies.add(graph.dependency)
            return False

        self.visit(add_to_dependencies, set())
        return dependencies

    @property
    def root_dependencies(self) -> set[Dependency]:
        roots: set[Dependency] = set()
        children: set[Dependency] = set()

        def add_to_dependencies(graph: DependencyGraph) -> bool:
            dependency = graph.dependency
            # if already encountered we're done
            if dependency in children:
                return False
            # if it's not a real file, then it's not a root
            if not dependency.path.is_file() and not is_a_notebook(dependency.path):
                children.add(dependency)
                return False
            # if it appears more than once then it can't be a root
            if dependency in roots:
                roots.remove(dependency)
                children.add(dependency)
                return False
            # if it has a 'real' parent, it's a child
            parent_graph = graph.parent
            while parent_graph is not None:
                dep = parent_graph.dependency
                if dep.path.is_file() or is_a_notebook(dep.path):
                    children.add(dependency)
                    return False
                parent_graph = parent_graph.parent
            # ok, it's a root
            roots.add(dependency)
            return False

        self.visit(add_to_dependencies, None)
        return roots

    @property
    def local_dependencies(self) -> set[Dependency]:
        return {child.dependency for child in self._dependencies.values()}

    @property
    def all_paths(self) -> set[Path]:
        # TODO: remove this public method, as it'll throw false positives
        # for package imports, like certifi. a WorkflowTask is also a dependency,
        # but it does not exist on a filesystem
        return {d.path for d in self.all_dependencies}

    def all_relative_names(self) -> set[str]:
        """This method is intended to simplify testing"""
        return self._relative_names(self.all_dependencies)

    def _relative_names(self, dependencies: set[Dependency]) -> set[str]:
        """This method is intended to simplify testing"""
        names = set[str]()
        for dependency in dependencies:
            for library_root in self._path_lookup.library_roots:
                if not dependency.path.is_relative_to(library_root):
                    continue
                relative_path = dependency.path.relative_to(library_root).as_posix()
                if relative_path == ".":
                    continue
                names.add(relative_path)
        return names

    def root_paths(self) -> set[Path]:
        return {d.path for d in self.root_dependencies}

    def root_relative_names(self) -> set[str]:
        """This method is intended to simplify testing"""
        return self._relative_names(self.root_dependencies)

    # when visit_node returns True it interrupts the visit
    def visit(self, visit_node: Callable[[DependencyGraph], bool | None], visited: set[Path] | None) -> bool:
        """provide visited set if you want to ensure nodes are only visited once"""
        if visited is not None:
            path = self.dependency.path
            if path in visited:
                return False
            visited.add(path)
        if visit_node(self):
            return True
        for dependency in self._dependencies.values():
            if dependency.visit(visit_node, visited):
                return True
        return False

    def new_dependency_graph_context(self):
        return DependencyGraphContext(
            parent=self, path_lookup=self._path_lookup, resolver=self._resolver, session_state=self._session_state
        )

    def _compute_route(self, root: Path, leaf: Path, visited: set[Path]) -> list[Dependency]:
        """given 2 files or notebooks root and leaf, compute the list of dependencies that must be traversed
        in order to lint the leaf in the context of its parents"""
        try:
            route = self._do_compute_route(root, leaf, visited)
            return self._trim_route(route)
        except ValueError:
            # we only compute routes on graphs so _compute_route can't fail
            # but we return an empty route in case it does
            return []

    def _do_compute_route(self, root: Path, leaf: Path, visited: set[Path]) -> list[Dependency]:
        maybe = self.locate_dependency(root)
        if not maybe.graph:
            logger.warning(f"Could not compute route because dependency {root} cannot be located")
            raise ValueError()
        route: list[Dependency] = []

        def do_compute_route(graph: DependencyGraph) -> bool:
            route.append(graph.dependency)
            for dependency in graph.local_dependencies:
                if dependency.path == leaf:
                    route.append(dependency)
                    return True
            for dependency in graph.local_dependencies:
                sub_route = self._do_compute_route(dependency.path, leaf, visited)
                if len(sub_route) > 0:
                    route.extend(sub_route)
                    return True
            route.pop()
            return False

        maybe.graph.visit(do_compute_route, visited)
        return route

    def _trim_route(self, dependencies: list[Dependency]) -> list[Dependency]:
        """don't inherit context if dependency is not a file/notebook, or it is loaded via dbutils.notebook.run or via import"""
        # filter out intermediate containers
        deps_with_source: list[Dependency] = []
        for dependency in dependencies:
            if dependency.path.is_file() or is_a_notebook(dependency.path):
                deps_with_source.append(dependency)
        # restart when not inheriting context
        for i, dependency in enumerate(deps_with_source):
            if dependency.inherits_context:
                continue
            return [dependency] + self._trim_route(deps_with_source[i + 1 :])
        return deps_with_source

    def build_inherited_tree(self, root: Path, leaf: Path) -> Tree | None:
        return self._build_inherited_context(root, leaf).tree

    def _build_inherited_context(self, root: Path, leaf: Path) -> InheritedContext:
        route = self._compute_route(root, leaf, set())
        return InheritedContext.from_route(self, self.path_lookup, route)

    def __repr__(self):
        return f"<DependencyGraph {self.dependency.path}>"


@dataclass
class DependencyGraphContext:
    parent: DependencyGraph
    path_lookup: PathLookup
    resolver: DependencyResolver
    session_state: CurrentSessionState


class Dependency(abc.ABC):

    def __init__(self, loader: DependencyLoader, path: Path, inherits_context=True):
        self._loader = loader
        self._path = path
        self._inherits_context = inherits_context

    @property
    def path(self) -> Path:
        return self._path

    @property
    def inherits_context(self):
        return self._inherits_context

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.path == other.path

    def load(self, path_lookup: PathLookup) -> SourceContainer | None:
        return self._loader.load_dependency(path_lookup, self)

    def __repr__(self):
        return f"Dependency<{self.path}>"


class SourceContainer(abc.ABC):

    @abc.abstractmethod
    def build_dependency_graph(self, parent: DependencyGraph) -> list[DependencyProblem]: ...

    def build_inherited_context(self, graph: DependencyGraph, child_path: Path) -> InheritedContext:
        raise ValueError(f"Building an inherited context from {type(self).__name__} is not supported!")


class DependencyLoader(abc.ABC):
    @abc.abstractmethod
    def load_dependency(self, path_lookup: PathLookup, dependency: Dependency) -> SourceContainer | None: ...


class WrappingLoader(DependencyLoader):

    def __init__(self, source_container: SourceContainer):
        self._source_container = source_container

    def load_dependency(self, path_lookup: PathLookup, dependency: Dependency) -> SourceContainer | None:
        return self._source_container

    def __repr__(self):
        return f"<WrappingLoader source_container={self._source_container}>"


class LibraryResolver(abc.ABC):
    @abc.abstractmethod
    def register_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        pass


class BaseNotebookResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_notebook(self, path_lookup: PathLookup, path: Path, inherit_context: bool) -> MaybeDependency: ...

    @staticmethod
    def _fail(code: str, message: str) -> MaybeDependency:
        return MaybeDependency(None, [DependencyProblem(code, message)])


class BaseImportResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_import(self, path_lookup: PathLookup, name: str) -> MaybeDependency:
        """resolve a simple or composite import name"""


class BaseFileResolver(abc.ABC):

    @abc.abstractmethod
    def resolve_file(self, path_lookup, path: Path) -> MaybeDependency:
        """locates a file"""


@dataclass
class MaybeDependency:
    dependency: Dependency | None
    problems: list[DependencyProblem]


class DependencyResolver:
    """the DependencyResolver is responsible for locating "stuff", mimicking the Databricks runtime behavior.
    There are specific resolution algorithms for import and for Notebooks (executed via %run or dbutils.notebook.run)
    so we're using 2 distinct resolvers for notebooks (instance) and imports (linked list of specialized sub-resolvers)
    resolving imports is affected by cwd and sys.paths, the latter being itself influenced by Python code
    we therefore need a PathLookup to convey these during import resolution
    """

    def __init__(
        self,
        library_resolver: LibraryResolver,
        notebook_resolver: BaseNotebookResolver,
        import_resolver: BaseImportResolver,
        file_resolver: BaseFileResolver,
        path_lookup: PathLookup,
    ):
        self._library_resolver = library_resolver
        self._notebook_resolver = notebook_resolver
        self._import_resolver = import_resolver
        self._file_resolver = file_resolver
        self._path_lookup = path_lookup

    def resolve_notebook(self, path_lookup: PathLookup, path: Path, inherit_context: bool) -> MaybeDependency:
        return self._notebook_resolver.resolve_notebook(path_lookup, path, inherit_context)

    def resolve_import(self, path_lookup: PathLookup, name: str) -> MaybeDependency:
        return self._import_resolver.resolve_import(path_lookup, name)

    def resolve_file(self, path_lookup: PathLookup, path: Path) -> MaybeDependency:
        return self._file_resolver.resolve_file(path_lookup, path)

    def register_library(self, path_lookup: PathLookup, *libraries: str) -> list[DependencyProblem]:
        return self._library_resolver.register_library(path_lookup, *libraries)

    def build_local_file_dependency_graph(self, path: Path, session_state: CurrentSessionState) -> MaybeGraph:
        """Builds a dependency graph starting from a file. This method is mainly intended for testing purposes.
        In case of problems, the paths in the problems will be relative to the starting path lookup."""
        maybe = self._file_resolver.resolve_file(self._path_lookup, path)
        if not maybe.dependency:
            return MaybeGraph(None, self._make_relative_paths(maybe.problems, path))
        graph = DependencyGraph(maybe.dependency, None, self, self._path_lookup, session_state)
        container = maybe.dependency.load(graph.path_lookup)
        if container is None:
            problem = DependencyProblem('cannot-load-file', f"Could not load file {path}")
            return MaybeGraph(None, [problem])
        problems = container.build_dependency_graph(graph)
        if problems:
            problems = self._make_relative_paths(problems, path)
        return MaybeGraph(graph, problems)

    def build_notebook_dependency_graph(self, path: Path, session_state: CurrentSessionState) -> MaybeGraph:
        """Builds a dependency graph starting from a notebook. This method is mainly intended for testing purposes.
        In case of problems, the paths in the problems will be relative to the starting path lookup."""
        # the notebook is the root of the graph, so there's no context to inherit
        maybe = self._notebook_resolver.resolve_notebook(self._path_lookup, path, inherit_context=False)
        if not maybe.dependency:
            return MaybeGraph(None, self._make_relative_paths(maybe.problems, path))
        graph = DependencyGraph(maybe.dependency, None, self, self._path_lookup, session_state)
        container = maybe.dependency.load(graph.path_lookup)
        if container is None:
            problem = DependencyProblem('cannot-load-notebook', f"Could not load notebook {path}")
            return MaybeGraph(None, [problem])
        problems = container.build_dependency_graph(graph)
        if problems:
            problems = self._make_relative_paths(problems, path)
        return MaybeGraph(graph, problems)

    def _make_relative_paths(self, problems: list[DependencyProblem], path: Path) -> list[DependencyProblem]:
        adjusted_problems = []
        for problem in problems:
            out_path = path if problem.is_path_missing() else problem.source_path
            if out_path.is_absolute() and out_path.is_relative_to(self._path_lookup.cwd):
                out_path = out_path.relative_to(self._path_lookup.cwd)
            adjusted_problems.append(problem.replace(source_path=out_path))
        return adjusted_problems

    def __repr__(self):
        return f"<DependencyResolver {self._notebook_resolver} {self._import_resolver} {self._file_resolver}, {self._path_lookup}>"


MISSING_SOURCE_PATH = "<MISSING_SOURCE_PATH>"


@dataclass
class DependencyProblem:
    code: str
    message: str
    source_path: Path = Path(MISSING_SOURCE_PATH)
    # Lines and columns are both 0-based: the first line is line 0.
    start_line: int = -1
    start_col: int = -1
    end_line: int = -1
    end_col: int = -1

    def is_path_missing(self):
        return self.source_path == Path(MISSING_SOURCE_PATH)

    def replace(
        self,
        code: str | None = None,
        message: str | None = None,
        source_path: Path | None = None,
        start_line: int | None = None,
        start_col: int | None = None,
        end_line: int | None = None,
        end_col: int | None = None,
    ) -> DependencyProblem:
        return DependencyProblem(
            code if code is not None else self.code,
            message if message is not None else self.message,
            source_path if source_path is not None else self.source_path,
            start_line if start_line is not None else self.start_line,
            start_col if start_col is not None else self.start_col,
            end_line if end_line is not None else self.end_line,
            end_col if end_col is not None else self.end_col,
        )

    def as_advisory(self) -> 'Advisory':
        return Advisory(
            code=self.code,
            message=self.message,
            start_line=self.start_line,
            start_col=self.start_col,
            end_line=self.end_line,
            end_col=self.end_col,
        )

    @staticmethod
    def from_node(code: str, message: str, node: NodeNG) -> DependencyProblem:
        # Astroid line numbers are 1-based.
        return DependencyProblem(
            code=code,
            message=message,
            start_line=(node.lineno or 1) - 1,
            start_col=node.col_offset,
            end_line=(node.end_lineno or 1) - 1,
            end_col=(node.end_col_offset or 0),
        )


@dataclass
class MaybeGraph:
    graph: DependencyGraph | None
    problems: list[DependencyProblem]

    @property
    def failed(self):
        return len(self.problems) > 0


class InheritedContext:

    @classmethod
    def from_route(cls, graph: DependencyGraph, path_lookup: PathLookup, route: list[Dependency]) -> InheritedContext:
        context = InheritedContext(None, False)
        for i, dependency in enumerate(route):
            if i >= len(route) - 1:
                break
            next_path = route[i + 1].path
            container = dependency.load(path_lookup)
            if container is None:
                logger.warning(f"Could not load content of {dependency.path}")
                return context
            local = container.build_inherited_context(graph, next_path)
            # only copy 'found' flag if this is the last parent
            context = context.append(local, i == len(route) - 2)
        return context.finalize()

    def __init__(self, tree: Tree | None, found: bool):
        self._tree = tree
        self._found = found

    @property
    def tree(self):
        return self._tree

    @property
    def found(self):
        return self._found

    def append(self, context: InheritedContext, copy_found: bool) -> InheritedContext:
        # we should never append to a found context
        if self.found:
            raise ValueError("Appending to an already resolved InheritedContext is illegal!")
        tree = context.tree
        found = copy_found and context.found
        if tree is None:
            return InheritedContext(self._tree, found)
        if self._tree is None:
            self._tree = Tree.new_module()
        self._tree.append_tree(context.tree)
        return InheritedContext(self._tree, found)

    def finalize(self) -> InheritedContext:
        # hacky stuff for aligning with Astroid's inference engine behavior
        # the engine checks line numbers to skip variables that are not in scope of the current frame
        # see https://github.com/pylint-dev/astroid/blob/5b665e7e760a7181625a24b3635e9fec7b174d87/astroid/filter_statements.py#L113
        # this is problematic when linting code fragments that refer to inherited code with unrelated line numbers
        # here we fool the engine by pretending that all nodes from context have negative line numbers
        if self._tree is None:
            return self
        tree = self._tree.renumber(-1)
        return InheritedContext(tree, self.found)


T = TypeVar("T")


class DependencyGraphWalker(abc.ABC, Generic[T]):

    def __init__(self, graph: DependencyGraph, walked_paths: set[Path], path_lookup: PathLookup):
        self._graph = graph
        self._walked_paths = walked_paths
        self._path_lookup = path_lookup

    def __iter__(self) -> Iterator[T]:
        for dependency in self._graph.root_dependencies:
            # the dependency is a root, so its path is the one to use
            # for computing lineage and building python global context
            root_path = dependency.path
            yield from self._iter_one(dependency, self._graph, root_path)

    def _iter_one(self, dependency: Dependency, graph: DependencyGraph, root_path: Path) -> Iterable[T]:
        if dependency.path in self._walked_paths:
            return
        self._walked_paths.add(dependency.path)
        self._log_walk_one(dependency)
        if dependency.path.is_file() or is_a_notebook(dependency.path):
            inherited_tree = graph.root.build_inherited_tree(root_path, dependency.path)
            path_lookup = self._path_lookup.change_directory(dependency.path.parent)
            yield from self._process_dependency(dependency, path_lookup, inherited_tree)
            maybe_graph = graph.locate_dependency(dependency.path)
            # missing graph problems have already been reported while building the graph
            if maybe_graph.graph:
                child_graph = maybe_graph.graph
                for child_dependency in child_graph.local_dependencies:
                    yield from self._iter_one(child_dependency, child_graph, root_path)

    def _log_walk_one(self, dependency: Dependency):
        logger.debug(f'Analyzing dependency: {dependency}')

    @abc.abstractmethod
    def _process_dependency(
        self, dependency: Dependency, path_lookup: PathLookup, inherited_tree: Tree | None
    ) -> Iterable[T]: ...



import dataclasses
import logging
import os
import re
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from functools import cached_property
from typing import ClassVar, Optional
from urllib.parse import urlparse

from databricks.labs.blueprint.installation import Installation
from databricks.labs.lsql import Row
from databricks.labs.lsql.backends import SqlBackend
from databricks.sdk import WorkspaceClient
from databricks.sdk.errors import NotFound
from databricks.sdk.service.catalog import ExternalLocationInfo
from databricks.sdk.dbutils import FileInfo
from databricks.labs.ucx.framework.crawlers import CrawlerBase
from databricks.labs.ucx.framework.utils import escape_sql_identifier
from databricks.labs.ucx.hive_metastore.tables import Table

logger = logging.getLogger(__name__)


_EXTERNAL_FILE_LOCATION_SCHEMES = ("s3", "s3a", "s3n", "gcs", "abfss")


@dataclass
class ExternalLocation:
    location: str
    table_count: int


@dataclass(unsafe_hash=True)
class Mount:
    name: str
    source: str


# Union with string is invalid syntax: "LocationTrie" | None
OptionalLocationTrie = Optional["LocationTrie"]  # pylint: disable=consider-alternative-union-syntax


@dataclass
class LocationTrie:
    """A trie datastructure to search locations.

    Used to find overlapping locations.
    """

    key: str = ""
    parent: OptionalLocationTrie = None
    children: dict[str, "LocationTrie"] = dataclasses.field(default_factory=dict)
    tables: list[Table] = dataclasses.field(default_factory=list)

    @cached_property
    def _path(self):
        """The path to traverse to get to the current node."""
        parts = []
        current = self
        while current:
            parts.append(current.key)
            current = current.parent
        return list(reversed(parts))[1:]

    @property
    def location(self):
        scheme, netloc, *path = self._path
        return f"{scheme}://{netloc}/{'/'.join(path)}"

    @staticmethod
    def _parse_location(location: str | None) -> list[str]:
        if not location:
            return []
        parse_result = urlparse(location)
        parts = [parse_result.scheme, parse_result.netloc]
        parts.extend(parse_result.path.strip("/").split("/"))
        return parts

    def insert(self, table: Table) -> None:
        current = self
        for part in self._parse_location(table.location):
            if part not in current.children:
                parent = current
                current = LocationTrie(part, parent)
                parent.children[part] = current
                continue
            current = current.children[part]
        current.tables.append(table)

    def find(self, table: Table) -> OptionalLocationTrie:
        current = self
        for part in self._parse_location(table.location):
            if part not in current.children:
                return None
            current = current.children[part]
        return current

    def is_valid(self) -> bool:
        """A valid location has a scheme and netloc; the path is optional."""
        if len(self._path) < 3:
            return False
        scheme, netloc, *_ = self._path
        return scheme in _EXTERNAL_FILE_LOCATION_SCHEMES and len(netloc) > 0

    def has_children(self):
        return len(self.children) > 0

    def __iter__(self):
        if self.is_valid():
            yield self
        for child in self.children.values():
            yield from child


class ExternalLocations(CrawlerBase[ExternalLocation]):
    _prefix_size: ClassVar[list[int]] = [1, 12]

    def __init__(self, ws: WorkspaceClient, sbe: SqlBackend, schema: str):
        super().__init__(sbe, "hive_metastore", schema, "external_locations", ExternalLocation)
        self._ws = ws

    def _external_locations(self, tables: list[Row], mounts) -> Iterable[ExternalLocation]:
        min_slash = 2
        external_locations: list[ExternalLocation] = []
        for table in tables:
            location = table.location
            if not location:
                continue
            if location.startswith("dbfs:/mnt"):
                location = self.resolve_mount(location, mounts)
            if (
                not location.startswith("dbfs")
                and (self._prefix_size[0] < location.find(":/") < self._prefix_size[1])
                and not location.startswith("jdbc")
            ):
                self._dbfs_locations(external_locations, location, min_slash)
            if location.startswith("jdbc"):
                self._add_jdbc_location(external_locations, location, table)
        return external_locations

    @staticmethod
    def resolve_mount(location, mounts):
        for mount in mounts:
            if location[5:].startswith(mount.name.lower()):
                location = location[5:].replace(mount.name, mount.source)
                break
        return location

    @staticmethod
    def _dbfs_locations(external_locations, location, min_slash):
        dupe = False
        loc = 0
        while loc < len(external_locations) and not dupe:
            common = (
                os.path.commonpath([external_locations[loc].location, os.path.dirname(location) + "/"]).replace(
                    ":/", "://"
                )
                + "/"
            )
            if common.count("/") > min_slash:
                table_count = external_locations[loc].table_count
                external_locations[loc] = ExternalLocation(common, table_count + 1)
                dupe = True
            loc += 1
        if not dupe:
            external_locations.append(ExternalLocation(os.path.dirname(location) + "/", 1))

    def _add_jdbc_location(self, external_locations, location, table):
        dupe = False
        pattern = r"(\w+)=(.*?)(?=\s*,|\s*\])"
        # Find all matches in the input string
        # Storage properties is of the format
        # "personalAccessToken=*********(redacted),
        #  httpPath=/sql/1.0/warehouses/65b52fb5bd86a7be, host=dbc-test1-aa11.cloud.databricks.com,
        #  dbtable=samples.nyctaxi.trips"
        matches = re.findall(pattern, table.storage_properties)
        # Create a dictionary from the matches
        result_dict = dict(matches)
        # Fetch the value of host from the newly created dict
        host = result_dict.get("host", "")
        port = result_dict.get("port", "")
        database = result_dict.get("database", "")
        httppath = result_dict.get("httpPath", "")
        provider = result_dict.get("provider", "")
        # currently supporting databricks and mysql external tables
        # add other jdbc types
        if "databricks" in location.lower():
            jdbc_location = f"jdbc:databricks://{host};httpPath={httppath}"
        elif "mysql" in location.lower():
            jdbc_location = f"jdbc:mysql://{host}:{port}/{database}"
        elif not provider == "":
            jdbc_location = f"jdbc:{provider.lower()}://{host}:{port}/{database}"
        else:
            jdbc_location = f"{location.lower()}/{host}:{port}/{database}"
        for ext_loc in external_locations:
            if ext_loc.location == jdbc_location:
                ext_loc.table_count += 1
                dupe = True
                break
        if not dupe:
            external_locations.append(ExternalLocation(jdbc_location, 1))

    def _external_location_list(self) -> Iterable[ExternalLocation]:
        tables_identifier = escape_sql_identifier(f"{self._catalog}.{self._schema}.tables")
        tables = list(
            self._backend.fetch(
                f"SELECT location, storage_properties FROM {tables_identifier} WHERE location IS NOT NULL"
            )
        )
        mounts = Mounts(self._backend, self._ws, self._schema).snapshot()
        return self._external_locations(list(tables), list(mounts))

    def snapshot(self) -> Iterable[ExternalLocation]:
        return self._snapshot(self._try_fetch, self._external_location_list)

    def _try_fetch(self) -> Iterable[ExternalLocation]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield ExternalLocation(*row)

    @staticmethod
    def _get_ext_location_definitions(missing_locations: list[ExternalLocation]) -> list:
        tf_script = []
        cnt = 1
        res_name = ""
        for loc in missing_locations:
            for scheme in _EXTERNAL_FILE_LOCATION_SCHEMES:
                prefix = f"{scheme}://"
                prefix_len = len(prefix)
                if not loc.location.startswith(prefix):
                    continue
                if prefix == "abfss://":
                    container_sep_loc = loc.location.index("@")
                    container_name = loc.location[prefix_len:container_sep_loc]
                    res_name = (
                        loc.location[container_sep_loc + 1 :]
                        .replace(".dfs.core.windows.net", "")
                        .rstrip("/")
                        .replace("/", "_")
                    )
                    res_name = f"{container_name}_{res_name}"
                else:
                    res_name = loc.location[prefix_len:].rstrip("/").replace("/", "_")
            if res_name == "":
                # if the cloud storage url doesn't match the above condition or incorrect (example wasb://)
                # dont generate tf script and ignore
                logger.warning(f"unsupported storage format {loc.location}")
                continue
            script = f'resource "databricks_external_location" "{res_name}" {{ \n'
            script += f'    name = "{res_name}"\n'
            script += f'    url  = "{loc.location.rstrip("/")}"\n'
            script += "    credential_name = databricks_storage_credential.<storage_credential_reference>.id\n"
            script += "}\n"
            tf_script.append(script)
            cnt += 1
        return tf_script

    def match_table_external_locations(self) -> tuple[dict[str, int], list[ExternalLocation]]:
        existing_locations = list(self._ws.external_locations.list())
        table_locations = self.snapshot()
        matching_locations: dict[str, int] = {}
        missing_locations = []
        for table_loc in table_locations:
            # external_location.list returns url without trailing "/" but ExternalLocation.snapshot
            # does so removing the trailing slash before comparing
            if not self._match_existing(table_loc, matching_locations, existing_locations):
                missing_locations.append(table_loc)
        return matching_locations, missing_locations

    @staticmethod
    def _match_existing(table_loc, matching_locations: dict[str, int], existing_locations: list[ExternalLocationInfo]):
        for uc_loc in existing_locations:
            if not uc_loc.url:
                continue
            if not uc_loc.name:
                continue
            uc_loc_path = uc_loc.url.lower()
            if uc_loc_path in table_loc.location.rstrip("/").lower():
                if uc_loc.name not in matching_locations:
                    matching_locations[uc_loc.name] = table_loc.table_count
                else:
                    matching_locations[uc_loc.name] = matching_locations[uc_loc.name] + table_loc.table_count
                return True
        return False

    def save_as_terraform_definitions_on_workspace(self, installation: Installation):
        matching_locations, missing_locations = self.match_table_external_locations()
        if len(matching_locations) > 0:
            logger.info("following external locations are already configured.")
            logger.info("sharing details of # tables that can be migrated for each location")
            for location, table_count in matching_locations.items():
                logger.info(f"{table_count} tables can be migrated using UC external location: {location}.")
        if len(missing_locations) > 0:
            logger.info("following external location need to be created.")
            for _ in missing_locations:
                logger.info(f"{_.table_count} tables can be migrated using external location {_.location}.")
            buffer = []
            for script in self._get_ext_location_definitions(missing_locations):
                buffer.append(script)
            return installation.upload('external_locations.tf', ("\n".join(buffer)).encode('utf8'))
        logger.info("no additional external location to be created.")
        return None


class Mounts(CrawlerBase[Mount]):
    def __init__(self, backend: SqlBackend, ws: WorkspaceClient, inventory_database: str):
        super().__init__(backend, "hive_metastore", inventory_database, "mounts", Mount)
        self._dbutils = ws.dbutils

    def _deduplicate_mounts(self, mounts: list) -> list:
        seen = set()
        deduplicated_mounts = []
        for obj in mounts:
            if "dbfsreserved" in obj.source.lower():
                obj_tuple = Mount("/Volume", obj.source)
            else:
                obj_tuple = Mount(obj.name, obj.source)
            if obj_tuple not in seen:
                seen.add(obj_tuple)
                deduplicated_mounts.append(obj)
        return deduplicated_mounts

    def inventorize_mounts(self):
        self._append_records(self._list_mounts())

    def _list_mounts(self) -> Iterable[Mount]:
        mounts = []
        for mount_point, source, _ in self._dbutils.fs.mounts():
            mounts.append(Mount(mount_point, source))
        return self._deduplicate_mounts(mounts)

    def snapshot(self) -> Iterable[Mount]:
        return self._snapshot(self._try_fetch, self._list_mounts)

    def _try_fetch(self) -> Iterable[Mount]:
        for row in self._fetch(f"SELECT * FROM {escape_sql_identifier(self.full_name)}"):
            yield Mount(*row)


@dataclass
class TableInMount:
    format: str
    is_partitioned: bool


class TablesInMounts(CrawlerBase[Table]):
    TABLE_IN_MOUNT_DB = "mounted_"

    def __init__(
        self,
        backend: SqlBackend,
        ws: WorkspaceClient,
        inventory_database: str,
        mc: Mounts,
        include_mounts: list[str] | None = None,
        exclude_paths_in_mount: list[str] | None = None,
        include_paths_in_mount: list[str] | None = None,
    ):
        super().__init__(backend, "hive_metastore", inventory_database, "tables", Table)
        self._dbutils = ws.dbutils
        self._mounts_crawler = mc
        self._include_mounts = include_mounts
        self._ws = ws
        self._include_paths_in_mount = include_paths_in_mount

        irrelevant_patterns = {'_SUCCESS', '_committed_', '_started_'}
        if exclude_paths_in_mount:
            irrelevant_patterns.update(exclude_paths_in_mount)
        self._fiter_paths = irrelevant_patterns

    def snapshot(self) -> list[Table]:
        logger.debug(f"[{self.full_name}] fetching {self._table} inventory")
        cached_results = []
        try:
            cached_results = list(self._try_load())
        except NotFound:
            pass
        table_paths = self._get_tables_paths_from_assessment(cached_results)
        logger.debug(f"[{self.full_name}] crawling new batch for {self._table}")
        loaded_records = list(self._crawl(table_paths))
        if len(cached_results) > 0:
            loaded_records = loaded_records + cached_results
        self._overwrite_records(loaded_records)
        return loaded_records

    def _try_load(self) -> Iterable[Table]:
        """Tries to load table information from the database or throws TABLE_OR_VIEW_NOT_FOUND error"""
        for row in self._fetch(
            f"SELECT * FROM {escape_sql_identifier(self.full_name)} WHERE NOT STARTSWITH(database, '{self.TABLE_IN_MOUNT_DB}')"
        ):
            yield Table(*row)

    def _get_tables_paths_from_assessment(self, loaded_records: Iterable[Table]) -> dict[str, str]:
        seen = {}
        for rec in loaded_records:
            if not rec.location:
                continue
            seen[rec.location] = rec.key
        return seen

    def _overwrite_records(self, items: Sequence[Table]):
        logger.debug(f"[{self.full_name}] found {len(items)} new records for {self._table}")
        self._backend.save_table(self.full_name, items, Table, mode="overwrite")

    def _crawl(self, table_paths_from_assessment: dict[str, str]):
        all_mounts = self._mounts_crawler.snapshot()
        all_tables = []
        for mount in all_mounts:
            if self._include_mounts and mount.name not in self._include_mounts:
                logger.info(f"Filtering mount {mount.name}")
                continue
            table_paths = {}
            if self._include_paths_in_mount:
                for path in self._include_paths_in_mount:
                    table_paths.update(self._find_delta_log_folders(path))
            else:
                table_paths = self._find_delta_log_folders(mount.name)

            for path, entry in table_paths.items():
                guess_table = os.path.basename(path)
                table_location = self._get_table_location(mount, path)

                # A table in mount may have already been identified by the assessment job because they're on the current workspace HMS
                # We filter those tables as we give better support to migrate those tables to UC
                if table_location in table_paths_from_assessment:
                    logger.info(
                        f"Path {path} is identified as a table in mount, but is present in current workspace as a registered table {table_paths_from_assessment[table_location]}"
                    )
                    continue
                if path in table_paths_from_assessment:
                    logger.info(
                        f"Path {path} is identified as a table in mount, but is present in current workspace as a registered table {table_paths_from_assessment[path]}"
                    )
                    continue
                table = Table(
                    catalog="hive_metastore",
                    database=f"{self.TABLE_IN_MOUNT_DB}{mount.name.replace('/mnt/', '').replace('/', '_')}",
                    name=guess_table,
                    object_type="EXTERNAL",
                    table_format=entry.format,
                    location=table_location,
                    is_partitioned=entry.is_partitioned,
                )
                all_tables.append(table)
        logger.info(f"Found a total of {len(all_tables)} tables in mount points")
        return all_tables

    def _get_table_location(self, mount: Mount, path: str):
        """
        There can be different cases for mounts:
            - Mount(name='/mnt/things/a', source='abfss://things@labsazurethings.dfs.core.windows.net/a')
            - Mount(name='/mnt/mount' source='abfss://container@dsss.net/')
            Both must return the complete source with a forward slash in the end
        """
        if mount.source.endswith("/"):
            return path.replace(f"dbfs:{mount.name}/", mount.source)
        return path.replace(f"dbfs:{mount.name}", mount.source)

    def _find_delta_log_folders(self, root_dir: str, delta_log_folders=None) -> dict:
        if delta_log_folders is None:
            delta_log_folders = {}
        logger.info(f"Listing {root_dir}")
        file_infos = self._dbutils.fs.ls(root_dir)
        for file_info in file_infos:
            if self._is_irrelevant(file_info.name) or file_info.path == root_dir:
                logger.debug(f"Path {file_info.path} is irrelevant")
                continue

            root_path = os.path.dirname(root_dir)
            previous_entry = delta_log_folders.get(root_path)
            table_in_mount = self._assess_path(file_info)

            if previous_entry:
                # Happens when first folder was _delta_log and next folders are partitioned folder
                if previous_entry.format == "DELTA" and self._is_partitioned(file_info.name):
                    delta_log_folders[root_path] = TableInMount(format=previous_entry.format, is_partitioned=True)
                # Happens when previous entries where partitioned folders and the current one is delta_log
                if previous_entry.is_partitioned and table_in_mount and table_in_mount.format == "DELTA":
                    delta_log_folders[root_path] = TableInMount(format=table_in_mount.format, is_partitioned=True)
                continue

            if self._is_partitioned(file_info.name):
                partition_format = self._find_partition_file_format(file_info.path)
                if partition_format:
                    delta_log_folders[root_path] = partition_format
                continue

            if not table_in_mount:
                self._find_delta_log_folders(file_info.path, delta_log_folders)
                continue

            delta_log_folders[root_path] = table_in_mount
        return delta_log_folders

    def _find_partition_file_format(self, root_dir: str) -> TableInMount | None:
        logger.info(f"Listing {root_dir}")
        file_infos = self._dbutils.fs.ls(root_dir)
        for file_info in file_infos:
            path_extension = self._assess_path(file_info)
            if path_extension:
                return TableInMount(format=path_extension.format, is_partitioned=True)
            if self._is_partitioned(file_info.name):
                return self._find_partition_file_format(file_info.path)
        return None

    def _assess_path(self, file_info: FileInfo) -> TableInMount | None:
        if file_info.name == "_delta_log/":
            logger.debug(f"Found delta table {file_info.path}")
            return TableInMount(format="DELTA", is_partitioned=False)
        if self._is_csv(file_info.name):
            logger.debug(f"Found csv {file_info.path}")
            return TableInMount(format="CSV", is_partitioned=False)
        if self._is_json(file_info.name):
            logger.debug(f"Found json {file_info.path}")
            return TableInMount(format="JSON", is_partitioned=False)
        if self._is_parquet(file_info.name):
            logger.debug(f"Found parquet {file_info.path}")
            return TableInMount(format="PARQUET", is_partitioned=False)
        return None

    def _is_partitioned(self, file_name: str) -> bool:
        return '=' in file_name

    def _is_parquet(self, file_name: str) -> bool:
        parquet_patterns = {'.parquet'}
        return any(pattern in file_name for pattern in parquet_patterns)

    def _is_csv(self, file_name: str) -> bool:
        csv_patterns = {'.csv'}
        return any(pattern in file_name for pattern in csv_patterns)

    def _is_json(self, file_name: str) -> bool:
        json_patterns = {'.json'}
        return any(pattern in file_name for pattern in json_patterns)

    def _is_irrelevant(self, file_name: str) -> bool:
        return any(pattern in file_name for pattern in self._fiter_paths)



from __future__ import annotations

import collections
import email
import json
import logging
import pkgutil
import re
import sys
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path

from databricks.labs.blueprint.entrypoint import get_logger

from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.source_code.base import CurrentSessionState
from databricks.labs.ucx.source_code.graph import DependencyProblem
from databricks.labs.ucx.source_code.linters.context import LinterContext
from databricks.labs.ucx.source_code.notebooks.sources import FileLinter
from databricks.labs.ucx.source_code.path_lookup import PathLookup

logger = logging.getLogger(__name__)

"""
Known libraries that are not in known.json

1) libraries with Python syntax that astroid cannot parse
tempo (error -> multiple exception types must be parenthesized)
chromadb (error -> 'Module' object has no attribute 'doc_node')

2) code that cannot be installed locally
dbruntime (error -> no pypi package)
horovod (error -> Failed to install temporary CMake. Please update your CMake to 3.13+ or set HOROVOD_CMAKE appropriately)
mediamix (requires a workspace, see https://d1r5llqwmkrl74.cloudfront.net/notebooks/CME/media-mix-modeling/index.html#media-mix-modeling_1.html)
mosaic (error -> print without parenthesis not supported by pip)
sat (requires a workspace, see https://github.com/databricks-industry-solutions/security-analysis-tool/blob/main/docs/setup.md)

3) code that cannot be located
utils
util
chedispy

4) code that's imported only once
"""


@dataclass
class Compatibility:
    known: bool
    problems: list[DependencyProblem]


@dataclass(unsafe_hash=True, frozen=True, eq=True, order=True)
class KnownProblem:
    code: str
    message: str

    def as_dict(self):
        return {'code': self.code, 'message': self.message}


UNKNOWN = Compatibility(False, [])
_DEFAULT_ENCODING = sys.getdefaultencoding()


class KnownList:
    def __init__(self):
        self._module_problems = collections.OrderedDict()
        self._library_problems = collections.defaultdict(list)
        known = self._get_known()
        for distribution_name, modules in known.items():
            specific_modules_first = sorted(modules.items(), key=lambda x: x[0], reverse=True)
            for module_ref, problems in specific_modules_first:
                module_problems = [DependencyProblem(**_) for _ in problems]
                self._module_problems[module_ref] = module_problems
                self._library_problems[distribution_name].extend(module_problems)
        for name in sys.stdlib_module_names:
            self._module_problems[name] = []

    @staticmethod
    def _get_known():
        module = __name__
        if __name__ == "__main__":  # code path for UCX developers invoking `make known`
            module = "databricks.labs.ucx.source_code.known"
        # load known.json from package data, because we may want to use zipapp packaging
        data = pkgutil.get_data(module, "known.json")
        return json.loads(data)

    def module_compatibility(self, name: str) -> Compatibility:
        if not name:
            return UNKNOWN
        for module, problems in self._module_problems.items():
            if not name.startswith(module):
                continue
            return Compatibility(True, problems)
        return UNKNOWN

    def distribution_compatibility(self, name: str) -> Compatibility:
        if not name:
            return UNKNOWN
        name = self._cleanup_name(name)
        problems = self._library_problems.get(name, None)
        if problems is None:
            return UNKNOWN
        return Compatibility(True, problems)

    @staticmethod
    def _cleanup_name(name):
        """parses the name to extract the library name, e.g. "numpy==1.21.0" -> "numpy",
        and "dist/databricks_labs_ucx-0.24.0-py3-none-any.whl" -> "databricks-labs-ucx"

        See https://pip.pypa.io/en/stable/reference/requirement-specifiers/#requirement-specifiers
        """
        _requirement_specifier_re = re.compile(r"([a-zA-Z0-9-]+)(?:[<>=].*)?")
        _wheel_name_re = re.compile(r"^([a-zA-Z0-9_]+)-.*\.whl$", re.MULTILINE)
        for matcher in (_wheel_name_re, _requirement_specifier_re):
            maybe_match = matcher.match(name)
            if not maybe_match:
                continue
            raw = maybe_match.group(1)
            return raw.replace('_', '-').lower()
        return name

    @classmethod
    def rebuild(cls, root: Path):
        """rebuild the known.json file by analyzing the source code of installed libraries. Invoked by `make known`."""
        path_lookup = PathLookup.from_sys_path(root)
        try:
            known_distributions = cls._get_known()
            logger.info("Scanning for newly installed distributions...")
        except FileNotFoundError:
            logger.info("No known distributions found; scanning all distributions...")
            known_distributions = {}
        updated_distributions = known_distributions.copy()
        for library_root in path_lookup.library_roots:
            for dist_info_folder in library_root.glob("*.dist-info"):
                cls._analyze_dist_info(dist_info_folder, updated_distributions, library_root)
        updated_distributions = dict(sorted(updated_distributions.items()))
        if known_distributions == updated_distributions:
            logger.info("No new distributions found.")
        else:
            known_json = Path(__file__).parent / "known.json"
            with known_json.open('w') as f:
                json.dump(updated_distributions, f, indent=2)
            logger.info(f"Updated known distributions: {known_json.relative_to(Path.cwd())}")

    @classmethod
    def _analyze_dist_info(cls, dist_info_folder, known_distributions, library_root):
        dist_info = DistInfo(dist_info_folder)
        if dist_info.name in known_distributions:
            logger.debug(f"Skipping distribution: {dist_info.name}")
            return
        logger.info(f"Processing distribution: {dist_info.name}")
        known_distributions[dist_info.name] = collections.OrderedDict()
        for module_path in dist_info.module_paths:
            if not module_path.is_file():
                continue
            if module_path.name in {'__main__.py', '__version__.py', '__about__.py'}:
                continue
            try:
                cls._analyze_file(known_distributions, library_root, dist_info, module_path)
            except RecursionError:
                logger.error(f"Recursion error in {module_path}")
                continue

    @classmethod
    def _analyze_file(cls, known_distributions, library_root, dist_info, module_path):
        empty_index = MigrationIndex([])
        relative_path = module_path.relative_to(library_root)
        module_ref = relative_path.as_posix().replace('/', '.')
        for suffix in ('.py', '.__init__'):
            if module_ref.endswith(suffix):
                module_ref = module_ref[: -len(suffix)]
        logger.info(f"Processing module: {module_ref}")
        session_state = CurrentSessionState()
        ctx = LinterContext(empty_index, session_state)
        linter = FileLinter(ctx, PathLookup.from_sys_path(module_path.parent), session_state, module_path)
        known_problems = set()
        for problem in linter.lint():
            known_problems.add(KnownProblem(problem.code, problem.message))
        problems = [_.as_dict() for _ in sorted(known_problems)]
        known_distributions[dist_info.name][module_ref] = problems

    def __repr__(self):
        modules = len(self._module_problems)
        libraries = len(self._library_problems)
        return f"<{self.__class__.__name__}: {modules} modules, {libraries} libraries>"


class DistInfo:
    """represents installed library in dist-info format
    see https://packaging.python.org/en/latest/specifications/binary-distribution-format/
    """

    def __init__(self, path: Path):
        self._path = path

    @cached_property
    def module_paths(self) -> list[Path]:
        files = []
        with Path(self._path, "RECORD").open(encoding=_DEFAULT_ENCODING) as f:
            for line in f.readlines():
                filename = line.split(',')[0]
                if not filename.endswith(".py"):
                    continue
                files.append(self._path.parent / filename)
        return files

    @cached_property
    def _metadata(self):
        with Path(self._path, "METADATA").open(encoding=_DEFAULT_ENCODING) as f:
            return email.message_from_file(f)

    @property
    def name(self):
        name = self._metadata.get('Name', 'unknown')
        return name.lower()

    @property
    def library_names(self) -> list[str]:
        names = []
        for requirement in self._metadata.get_all('Requires-Dist', []):
            library = self._extract_library_name_from_requires_dist(requirement)
            names.append(library)
        return names

    @staticmethod
    def _extract_library_name_from_requires_dist(requirement: str) -> str:
        delimiters = {' ', '@', '<', '>', ';'}
        for i, char in enumerate(requirement):
            if char in delimiters:
                return requirement[:i]
        return requirement

    def __repr__(self):
        return f"<DistInfoPackage {self._path}>"


if __name__ == "__main__":
    logger = get_logger(__file__)  # this only works for __main__
    KnownList.rebuild(Path.cwd())



import abc
import logging
from datetime import timedelta
from functools import cached_property
from pathlib import Path

from databricks.labs.blueprint.installation import Installation
from databricks.labs.blueprint.installer import InstallState
from databricks.labs.blueprint.tui import Prompts
from databricks.labs.blueprint.wheels import ProductInfo, WheelsV2
from databricks.labs.lsql.backends import SqlBackend

from databricks.labs.ucx.recon.data_comparator import StandardDataComparator
from databricks.labs.ucx.recon.data_profiler import StandardDataProfiler
from databricks.labs.ucx.recon.metadata_retriever import DatabricksTableMetadataRetriever
from databricks.labs.ucx.recon.migration_recon import MigrationRecon
from databricks.labs.ucx.recon.schema_comparator import StandardSchemaComparator
from databricks.labs.ucx.source_code.python_libraries import PythonLibraryResolver
from databricks.sdk import AccountClient, WorkspaceClient, core
from databricks.sdk.errors import ResourceDoesNotExist
from databricks.sdk.service import sql

from databricks.labs.ucx.account.workspaces import WorkspaceInfo
from databricks.labs.ucx.assessment.azure import AzureServicePrincipalCrawler
from databricks.labs.ucx.aws.credentials import CredentialManager
from databricks.labs.ucx.config import WorkspaceConfig
from databricks.labs.ucx.hive_metastore import ExternalLocations, Mounts, TablesCrawler
from databricks.labs.ucx.hive_metastore.catalog_schema import CatalogSchema
from databricks.labs.ucx.hive_metastore.grants import (
    AzureACL,
    GrantsCrawler,
    PrincipalACL,
    AwsACL,
    MigrateGrants,
    ACLMigrator,
)
from databricks.labs.ucx.hive_metastore.mapping import TableMapping
from databricks.labs.ucx.hive_metastore.migration_status import MigrationIndex
from databricks.labs.ucx.hive_metastore.table_migrate import (
    MigrationStatusRefresher,
    TablesMigrator,
)
from databricks.labs.ucx.hive_metastore.table_move import TableMove
from databricks.labs.ucx.hive_metastore.udfs import UdfsCrawler
from databricks.labs.ucx.hive_metastore.verification import VerifyHasMetastore
from databricks.labs.ucx.installer.workflows import DeployedWorkflows
from databricks.labs.ucx.source_code.jobs import WorkflowLinter
from databricks.labs.ucx.source_code.notebooks.loaders import (
    NotebookResolver,
    NotebookLoader,
)
from databricks.labs.ucx.source_code.linters.files import FileLoader, FolderLoader, ImportFileResolver
from databricks.labs.ucx.source_code.path_lookup import PathLookup
from databricks.labs.ucx.source_code.graph import DependencyResolver
from databricks.labs.ucx.source_code.known import KnownList
from databricks.labs.ucx.source_code.redash import Redash
from databricks.labs.ucx.workspace_access import generic, redash
from databricks.labs.ucx.workspace_access.groups import GroupManager
from databricks.labs.ucx.workspace_access.manager import PermissionManager
from databricks.labs.ucx.workspace_access.scim import ScimSupport
from databricks.labs.ucx.workspace_access.secrets import SecretScopesSupport
from databricks.labs.ucx.workspace_access.tacl import TableAclSupport

# "Service Factories" would always have a lot of public methods.
# This is because they are responsible for creating objects that are
# used throughout the application. That being said, we'll do best
# effort of splitting the instances between Global, Runtime,
# Workspace CLI, and Account CLI contexts.
# pylint: disable=too-many-public-methods

logger = logging.getLogger(__name__)


class GlobalContext(abc.ABC):
    def __init__(self, named_parameters: dict[str, str] | None = None):
        if not named_parameters:
            named_parameters = {}
        self._named_parameters = named_parameters

    def replace(self, **kwargs):
        """Replace cached properties for unit testing purposes."""
        for key, value in kwargs.items():
            self.__dict__[key] = value
        return self

    @cached_property
    def workspace_client(self) -> WorkspaceClient:
        raise ValueError("Workspace client not set")

    @cached_property
    def sql_backend(self) -> SqlBackend:
        raise ValueError("SQL backend not set")

    @cached_property
    def account_client(self) -> AccountClient:
        raise ValueError("Account client not set")

    @cached_property
    def named_parameters(self) -> dict[str, str]:
        return self._named_parameters

    @cached_property
    def product_info(self):
        return ProductInfo.from_class(WorkspaceConfig)

    @cached_property
    def installation(self):
        return Installation.current(self.workspace_client, self.product_info.product_name())

    @cached_property
    def config(self) -> WorkspaceConfig:
        return self.installation.load(WorkspaceConfig)

    @cached_property
    def connect_config(self) -> core.Config:
        return self.workspace_client.config

    @cached_property
    def is_azure(self) -> bool:
        if self.is_aws:
            return False
        return self.connect_config.is_azure

    @cached_property
    def is_aws(self) -> bool:
        return self.connect_config.is_aws

    @cached_property
    def is_gcp(self) -> bool:
        return not self.is_aws and not self.is_azure

    @cached_property
    def inventory_database(self) -> str:
        return self.config.inventory_database

    @cached_property
    def workspace_listing(self):
        return generic.WorkspaceListing(
            self.workspace_client,
            self.sql_backend,
            self.inventory_database,
            self.config.num_threads,
            self.config.workspace_start_path,
        )

    @cached_property
    def generic_permissions_support(self):
        models_listing = generic.models_listing(self.workspace_client, self.config.num_threads)
        acl_listing = [
            generic.Listing(self.workspace_client.clusters.list, "cluster_id", "clusters"),
            generic.Listing(self.workspace_client.cluster_policies.list, "policy_id", "cluster-policies"),
            generic.Listing(self.workspace_client.instance_pools.list, "instance_pool_id", "instance-pools"),
            generic.Listing(self.workspace_client.warehouses.list, "id", "sql/warehouses"),
            generic.Listing(self.workspace_client.jobs.list, "job_id", "jobs"),
            generic.Listing(self.workspace_client.pipelines.list_pipelines, "pipeline_id", "pipelines"),
            generic.Listing(self.workspace_client.serving_endpoints.list, "id", "serving-endpoints"),
            generic.Listing(generic.experiments_listing(self.workspace_client), "experiment_id", "experiments"),
            generic.Listing(models_listing, "id", "registered-models"),
            generic.Listing(generic.models_root_page, "object_id", "registered-models"),
            generic.Listing(generic.tokens_and_passwords, "object_id", "authorization"),
            generic.Listing(generic.feature_store_listing(self.workspace_client), "object_id", "feature-tables"),
            generic.Listing(generic.feature_tables_root_page, "object_id", "feature-tables"),
            self.workspace_listing,
        ]
        return generic.GenericPermissionsSupport(
            self.workspace_client,
            acl_listing,
            include_object_permissions=self.config.include_object_permissions,
        )

    @cached_property
    def redash_permissions_support(self):
        acl_listing = [
            redash.Listing(self.workspace_client.alerts.list, sql.ObjectTypePlural.ALERTS),
            redash.Listing(self.workspace_client.dashboards.list, sql.ObjectTypePlural.DASHBOARDS),
            redash.Listing(self.workspace_client.queries.list, sql.ObjectTypePlural.QUERIES),
        ]
        return redash.RedashPermissionsSupport(
            self.workspace_client,
            acl_listing,
            include_object_permissions=self.config.include_object_permissions,
        )

    @cached_property
    def scim_entitlements_support(self):
        return ScimSupport(self.workspace_client, include_object_permissions=self.config.include_object_permissions)

    @cached_property
    def secret_scope_acl_support(self):
        return SecretScopesSupport(
            self.workspace_client, include_object_permissions=self.config.include_object_permissions
        )

    @cached_property
    def legacy_table_acl_support(self):
        return TableAclSupport(
            self.grants_crawler,
            self.sql_backend,
            include_object_permissions=self.config.include_object_permissions,
        )

    @cached_property
    def permission_manager(self):
        return PermissionManager(
            self.sql_backend,
            self.inventory_database,
            [
                self.generic_permissions_support,
                self.redash_permissions_support,
                self.secret_scope_acl_support,
                self.scim_entitlements_support,
                self.legacy_table_acl_support,
            ],
        )

    @cached_property
    def group_manager(self):
        return GroupManager(
            self.sql_backend,
            self.workspace_client,
            self.inventory_database,
            self.config.include_group_names,
            self.config.renamed_group_prefix,
            workspace_group_regex=self.config.workspace_group_regex,
            workspace_group_replace=self.config.workspace_group_replace,
            account_group_regex=self.config.account_group_regex,
            external_id_match=self.config.group_match_by_external_id,
        )

    @cached_property
    def grants_crawler(self):
        return GrantsCrawler(self.tables_crawler, self.udfs_crawler, self.config.include_databases)

    @cached_property
    def udfs_crawler(self):
        return UdfsCrawler(self.sql_backend, self.inventory_database, self.config.include_databases)

    @cached_property
    def tables_crawler(self):
        return TablesCrawler(self.sql_backend, self.inventory_database, self.config.include_databases)

    @cached_property
    def tables_migrator(self):
        return TablesMigrator(
            self.tables_crawler,
            self.workspace_client,
            self.sql_backend,
            self.table_mapping,
            self.migration_status_refresher,
            self.migrate_grants,
        )

    @cached_property
    def acl_migrator(self):
        return ACLMigrator(
            self.tables_crawler,
            self.workspace_info,
            self.migration_status_refresher,
            self.migrate_grants,
        )

    @cached_property
    def migrate_grants(self):
        grant_loaders = [
            self.grants_crawler.snapshot,
            self.principal_acl.get_interactive_cluster_grants,
        ]
        return MigrateGrants(
            self.sql_backend,
            self.group_manager,
            grant_loaders,
        )

    @cached_property
    def table_move(self):
        return TableMove(self.workspace_client, self.sql_backend)

    @cached_property
    def mounts_crawler(self):
        return Mounts(self.sql_backend, self.workspace_client, self.inventory_database)

    @cached_property
    def azure_service_principal_crawler(self):
        return AzureServicePrincipalCrawler(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def external_locations(self):
        return ExternalLocations(self.workspace_client, self.sql_backend, self.inventory_database)

    @cached_property
    def azure_acl(self):
        return AzureACL(
            self.workspace_client,
            self.sql_backend,
            self.azure_service_principal_crawler,
            self.installation,
        )

    @cached_property
    def aws_acl(self):
        return AwsACL(
            self.workspace_client,
            self.sql_backend,
            self.installation,
        )

    @cached_property
    def principal_locations(self):
        eligible_locations = {}
        try:
            if self.is_azure:
                eligible_locations = self.azure_acl.get_eligible_locations_principals()
            if self.is_aws:
                eligible_locations = self.aws_acl.get_eligible_locations_principals()
            if self.is_gcp:
                raise NotImplementedError("Not implemented for GCP.")
        except ResourceDoesNotExist:
            pass
        return eligible_locations

    @cached_property
    def principal_acl(self):
        return PrincipalACL(
            self.workspace_client,
            self.sql_backend,
            self.installation,
            self.tables_crawler,
            self.mounts_crawler,
            self.principal_locations,
        )

    @cached_property
    def migration_status_refresher(self):
        return MigrationStatusRefresher(
            self.workspace_client,
            self.sql_backend,
            self.inventory_database,
            self.tables_crawler,
        )

    @cached_property
    def iam_credential_manager(self):
        return CredentialManager(self.workspace_client)

    @cached_property
    def table_mapping(self):
        return TableMapping(self.installation, self.workspace_client, self.sql_backend)

    @cached_property
    def catalog_schema(self):
        return CatalogSchema(self.workspace_client, self.table_mapping, self.principal_acl, self.sql_backend)

    @cached_property
    def verify_timeout(self):
        return timedelta(minutes=2)

    @cached_property
    def wheels(self):
        return WheelsV2(self.installation, self.product_info)

    @cached_property
    def install_state(self):
        return InstallState.from_installation(self.installation)

    @cached_property
    def deployed_workflows(self):
        return DeployedWorkflows(self.workspace_client, self.install_state, self.verify_timeout)

    @cached_property
    def workspace_info(self):
        return WorkspaceInfo(self.installation, self.workspace_client)

    @cached_property
    def verify_has_metastore(self):
        return VerifyHasMetastore(self.workspace_client)

    @cached_property
    def pip_resolver(self):
        return PythonLibraryResolver(self.allow_list)

    @cached_property
    def notebook_loader(self) -> NotebookLoader:
        return NotebookLoader()

    @cached_property
    def notebook_resolver(self):
        return NotebookResolver(self.notebook_loader)

    @cached_property
    def site_packages_path(self):
        lookup = self.path_lookup
        return next(path for path in lookup.library_roots if "site-packages" in path.as_posix())

    @cached_property
    def path_lookup(self):
        # TODO find a solution to enable a different cwd per job/task (maybe it's not necessary or possible?)
        return PathLookup.from_sys_path(Path.cwd())

    @cached_property
    def file_loader(self):
        return FileLoader()

    @cached_property
    def folder_loader(self):
        return FolderLoader(self.notebook_loader, self.file_loader)

    @cached_property
    def allow_list(self):
        return KnownList()

    @cached_property
    def file_resolver(self):
        return ImportFileResolver(self.file_loader, self.allow_list)

    @cached_property
    def dependency_resolver(self):
        return DependencyResolver(
            self.pip_resolver, self.notebook_resolver, self.file_resolver, self.file_resolver, self.path_lookup
        )

    @cached_property
    def workflow_linter(self):
        return WorkflowLinter(
            self.workspace_client,
            self.dependency_resolver,
            self.path_lookup,
            MigrationIndex([]),  # TODO: bring back self.tables_migrator.index()
            self.config.include_job_ids,
        )

    @cached_property
    def redash(self):
        return Redash(
            self.migration_status_refresher.index(),
            self.workspace_client,
            self.installation,
        )

    @cached_property
    def metadata_retriever(self):
        return DatabricksTableMetadataRetriever(self.sql_backend)

    @cached_property
    def schema_comparator(self):
        return StandardSchemaComparator(self.metadata_retriever)

    @cached_property
    def data_profiler(self):
        return StandardDataProfiler(self.sql_backend, self.metadata_retriever)

    @cached_property
    def data_comparator(self):
        return StandardDataComparator(self.sql_backend, self.data_profiler)

    @cached_property
    def migration_recon(self):
        return MigrationRecon(
            self.sql_backend,
            self.inventory_database,
            self.migration_status_refresher,
            self.table_mapping,
            self.schema_comparator,
            self.data_comparator,
            self.config.recon_tolerance_percent,
        )


class CliContext(GlobalContext, abc.ABC):
    @cached_property
    def prompts(self) -> Prompts:
        return Prompts()
